from typing import Union, List, Dict, Tuple
import psutil
import GPUtil
import numpy as np
import torch as t
import torch.nn as nn


class ModelSizeEstimator:
    """
    Size estimator for pytorch modules.
    """
    def __init__(self,
                 model: nn.Module,
                 size_multiplier=2):
        """
        Estimates the size of PyTorch models in memory.

        Note:
            This estimator can only estimate the total size of parameters and
            buffers. Therefore we need to multiply the raw estimated size with
            a correction coefficient to reserve enough space for models.

        Args:
            model: Model to be estimated.
            size_multiplier: Model estimated size will be
                multiplied with this value, to ensure enough space
                will be reserved to contain your model and inputs.
        """
        self.model = model
        self.size_multiplier = size_multiplier
        self.sizes = {}
        self.dtype_sizes = {}

    def get_parameter_sizes(self) -> float:
        """Get sizes of all parameters in ``model`` in mega bytes."""
        sizes, dtype_sizes = [], []

        for param in self.model.parameters():
            sizes.append(np.array(param.shape))
            dtype_sizes.append(self._get_dtype_in_bytes(param.dtype))
        self.sizes["param"] = sizes
        self.dtype_sizes["param"] = dtype_sizes
        return float(
            np.sum(np.array([np.prod(s) for s in self.sizes["param"]]) *
                   np.array(self.dtype_sizes["param"]))
        ) / (1024 ** 2)

    def get_buffer_sizes(self) -> float:
        """Get sizes of all buffers in ``model`` in mega bytes."""
        sizes, dtype_sizes = [], []

        for buffer in self.model.buffers():
            sizes.append(np.array(buffer.shape))
            dtype_sizes.append(self._get_dtype_in_bytes(buffer.dtype))
        self.sizes["buffer"] = sizes
        self.dtype_sizes["buffer"] = dtype_sizes
        return float(
            np.sum(np.array([np.prod(s) for s in self.sizes["buffer"]]) *
                   np.array(self.dtype_sizes["buffer"]))
        ) / (1024 ** 2)

    def estimate_size(self):
        """Estimate model size in memory in megabytes."""
        total = self.get_parameter_sizes() + self.get_buffer_sizes()
        return total * self.size_multiplier

    @staticmethod
    def _get_dtype_in_bytes(dtype: t.dtype):
        if dtype in (t.int8, t.uint8, t.bool):
            return 1
        elif dtype in (t.int16, t.float16, t.short, t.half):
            return 2
        elif dtype in (t.int32, t.float32, t.int, t.float):
            return 4
        elif dtype in (t.int64, t.float64, t.long, t.double, t.complex32):
            return 8
        elif dtype in (t.complex64,):
            return 16
        elif dtype in (t.complex128,):
            return 32
        else:
            raise ValueError("Invalid data type.")


class ModelAssigner:
    """
    Assigner for pytorch modules.
    """
    def __init__(self,
                 models: List[nn.Module],
                 model_connection: Dict[Tuple[int, int], int],
                 devices: List[Union[t.device, str]] = None,
                 model_size_multiplier=2,
                 max_mem_ratio=0.5,
                 cpu_weight=1,
                 distance_weight=5,
                 size_balance_weight=1e-3,
                 complexity_balance_weight=5,
                 entropy_weight=1,
                 iterations=500,
                 update_rate=0.01,
                 gpu_gpu_distance=1,
                 cpu_gpu_distance=10,
                 move_models=True):
        """
        Assign models to different devices. In the scope of a single process.
        Assigner assumes all GPUs have the **same processing power**.

        Assignment is based on four aspects:

        1. Distance and model connections. Connection is usually indicated
           by the amount of data transimitted between two models.
        2. Compute complexity.
        3. Model size.
        4. Entropy.

        Four aspects are controlled by four weights:

        1. ``distance_weight``, assigner will try to reduce the total
           ``distance * connection`` if this weight is larger.
        2. ``size_balance_weight``, this weight controls the total memory
           space used on a single device, only works if total assigned
           memory of models exceeds allowed device memory size
           (internally it uses a relu activation), the larger,
           the tighter and more restricted the fit.
        3. ``complexity_balance_weight``, this weights balance the model
           computation cost across devices, assigner will try to even
           the ``computation cost / compute power`` ratio for each device
           if this weight is larger.
        4. ``entropy_weight``, this weight minimize the uncertainty of
           model placement probability, so ``model i`` will have a close to 1
           probability of locating on some ``device j`` if this weight is
           larger.

        Assignment uses gradient descent to compute the probability matrix
        of each ``model i`` locating on each available ``device j``.

        See Also:
            :class:`.ModelSizeEstimator`

        Args:
            models: Models to assign.
            model_connection: Connection weight between modules.
            devices: Available devices.
            model_size_multiplier: Size multiplier of models, used to reserve
                enough space for models,
            max_mem_ratio: Maximum percent of memory allowed.
            cpu_weight: Weight of cpu. Relative to the computing power of one
                GPU.
            distance_weight: Weight of distance.
            size_balance_weight: Weight of size balance.
            complexity_balance_weight: Weight of complexity balance.
            entropy_weight: Weight of entropy.
            iterations: Number of optimization iterations.
            update_rate: Learning rate of the adam optimizer.
            gpu_gpu_distance: Estimated distance cost between gpu-gpu.
            cpu_gpu_distance: Estimated distance cost between cpu-gpu.
            move_models: Wether to automtically move the models after
                assignment.
        """
        if devices is None:
            devices = [t.device(type="cuda", index=i)
                       for i in GPUtil.getAvailable(order="load")]
        else:
            devices = [t.device(d) for d in devices]
            available_devices = [t.device(type="cuda", index=i)
                                 for i in GPUtil.getAvailable(order="load")]
            used_devices = []
            for dev in devices:
                if dev.type == "cuda" and dev not in available_devices:
                    print("Warning: device {} not available, removed."
                          .format(dev))
                else:
                    used_devices.append(dev)
            devices = used_devices

        if not devices:
            devices = [t.device("cpu")]

        print("Using these devices: {}".format(devices))

        sizes = [ModelSizeEstimator(model, model_size_multiplier)
                 .estimate_size()
                 for model in models]
        device_size_capacity = []
        device_complexity_capacity = []

        gpus = GPUtil.getGPUs()
        for dev in devices:
            if dev.type == "cpu":
                device_size_capacity.append(
                    int(psutil.virtual_memory().available / 1024 ** 2) *
                    max_mem_ratio
                )
                device_complexity_capacity.append(cpu_weight)
            elif dev.type == "cuda":
                device_size_capacity.append(
                    gpus[dev.index].memoryFree * max_mem_ratio
                )
                device_complexity_capacity.append(1 - gpus[dev.index].load)

        if np.sum(np.array(sizes)) > np.sum(device_size_capacity):
            raise RuntimeError("Estimated model will use {:.2f} MB, "
                               "but only have {:.2f} MB allowed memory "
                               "in total."
                               .format(np.sum(np.array(sizes)),
                                       np.sum(device_size_capacity)))

        # assign model to devices
        # using heuristic and gradient decent
        device_num = len(devices)
        model_num = len(models)

        # Important, the placement probability matrix! this matrix
        # describes the probability of placement of:
        # model i on device j
        placement = t.randn([model_num, device_num], requires_grad=True)

        optimizer = t.optim.Adam([placement], lr=update_rate)
        model_size = t.tensor(sizes, dtype=t.float).view([1, model_num])
        size_capacity = (t.tensor(device_size_capacity, dtype=t.float)
                         .view([1, device_num]))
        model_complexity = model_size

        # complexity_capacity is basically the estimated computing power
        # of devices.
        complexity_capacity = t.tensor(device_complexity_capacity,
                                       dtype=t.float).view([1, device_num])

        # model connection indicates the amount of data transmitted between
        # each pair of models, a weighted adjacency matrix.
        model_conn = t.zeros([model_num, model_num])

        for direction, conn in model_connection.items():
            model_conn[direction[0], direction[1]] = conn

        # device distance matrix
        device_distance = t.zeros([device_num, device_num])
        for i in range(device_num):
            for j in range(i):
                if (devices[i].type == "cpu" and
                        devices[j].type == "cuda" or
                        devices[i].type == "cuda" and
                        devices[j].type == "cpu"):
                    device_distance[i, j] = device_distance[j, i] = \
                        cpu_gpu_distance
                elif (devices[i].type == "cuda" and
                      devices[j].type == "cuda" and
                      devices[i].index != devices[j].index):
                    device_distance[i, j] = device_distance[j, i] = \
                        gpu_gpu_distance

        # optimize
        for _ in range(iterations):
            self.optimize_placement(optimizer, placement,
                                    model_size, size_capacity,
                                    model_complexity, complexity_capacity,
                                    model_conn, device_distance,
                                    distance_weight,
                                    size_balance_weight,
                                    complexity_balance_weight,
                                    entropy_weight)
        self._assignment = [devices[d]
                            for d
                            in t.argmax(placement, dim=1).tolist()]
        if move_models:
            for model, ass_device in zip(models, self._assignment):
                model.to(ass_device)

    @property
    def assignment(self):
        """
        List[t.device]:
            Assigned devices for each model in your model list.
        """
        return self._assignment

    @staticmethod
    def optimize_placement(optimizer,
                           placement: t.Tensor,
                           model_size: t.Tensor,
                           size_capacity: t.Tensor,
                           model_complexity: t.Tensor,
                           complexity_capacity: t.Tensor,
                           model_connection: t.Tensor,
                           device_distance: t.Tensor,
                           distance_weight: float,
                           size_balance_weight: float,
                           complexity_balance_weight: float,
                           entropy_weight: float):
        """
        Suppose there are n models to place and m devices available.

        Args:
            optimizer: optimizer of placement
            placement: shape ``[n, m]``
            model_size: shape ``[1, n]``
            size_capacity: shape ``[1, m]``
            model_complexity: shape ``[1, n]``
            complexity_capacity: shape ``[1, m]``
            model_connection: shape ``[n, n]``
            device_distance: shape ``[m, m]``
            distance_weight: Weight of distance.
            size_balance_weight: Weight of size balance.
            complexity_balance_weight: Weight of complexity balance.
            entropy_weight: weight of entropy.
        """
        placement = t.softmax(placement, dim=-1)
        model_num = placement.shape[0]

        model_distance = t.einsum("ij,mn,jn->im",
                                  placement, placement, device_distance)
        model_distance[np.arange(model_num), np.arange(model_num)] = 0
        connection_cost = model_connection * model_distance

        # sum(model size) < capacity
        size_match_cost = t.relu(t.einsum("ij,jk->ik", model_size, placement) -
                                 size_capacity)

        # match computing power percent
        norm_model_cmplx = model_complexity / t.sum(model_complexity)
        norm_cmplx_capacity = complexity_capacity / t.sum(complexity_capacity)
        cmplx_match_cost = (t.einsum("ij,jk->ik", norm_model_cmplx, placement) -
                            norm_cmplx_capacity) ** 2

        # entropy loss, prevent placement probability diffuse over devices
        entropy_cost = placement * placement.log()
        tmp = t.zeros_like(placement)
        entropy_cost = -t.where(placement > 0, entropy_cost, tmp).sum(dim=-1)

        total_cost = (t.mean(connection_cost) * distance_weight +
                      t.mean(size_match_cost) * size_balance_weight +
                      t.mean(cmplx_match_cost) * complexity_balance_weight +
                      t.mean(entropy_cost) * entropy_weight)

        optimizer.zero_grad()
        total_cost.backward()
        optimizer.step()



from .base import TorchFramework

from .dqn import DQN
from .dqn_per import DQNPer
from .rainbow import RAINBOW

from .ddpg import DDPG
from .hddpg import HDDPG
from .td3 import TD3
from .ddpg_per import DDPGPer

from .a2c import A2C
from .a3c import A3C
from .ppo import PPO
from .sac import SAC

from .apex import DQNApex, DDPGApex
from .impala import IMPALA

from .maddpg import MADDPG

__all__ = [
    "TorchFramework",
    "DQN", "DQNPer", "RAINBOW",
    "DDPG", "HDDPG", "TD3", "DDPGPer",
    "A2C", "A3C", "PPO", "SAC",
    "DQNApex", "DDPGApex", "IMPALA",
    "MADDPG",
]



from os import path
import re
import setuptools

root = path.abspath(path.dirname(__file__))

with open(path.join(root, "machin", "__init__.py")) as f:
    version = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)

with open("README.md", mode="r", encoding="utf8") as desc:
    long_description = desc.read()

setuptools.setup(
    name="machin",
    version=version,
    description="Reinforcement learning library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iffiX/machin",
    author="Iffi",
    author_email="iffi@mail.beyond-infinity.com",
    license="MIT",
    python_requires=">=3.5",
    packages=setuptools.find_packages(
        exclude=["test", "test.*", "examples", "examples.*",
                 "docs", "docs.*"]
    ),
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",

        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        "gym",
        "psutil",
        "numpy",
        "torch>=1.5.0",
        "torchviz",
        "moviepy",
        "matplotlib",
        "colorlog",
        "dill",
        "GPUtil",
        "Pillow",
        "tensorboardX",
    ]
)



from typing import Union, Dict, List, Tuple, Any
from ..transition import Transition
from .buffer import Buffer
import torch as t
import numpy as np


class WeightTree:
    """
    Sum weight tree data structure.
    """
    def __init__(self, size):
        """
        Initialize a weight tree.

        Note:
            Weights must be positive.

        Note:
            Weight tree is stored as a flattened, full binary tree in a
            ``np.ndarray``. The lowest level of leaves comes first, the
            highest root node is stored at last.

            Example:

            Tree with weights: ``[[1, 2, 3, 4], [3, 7], [11]]``

            will be stored as: ``[1, 2, 3, 4, 3, 7, 11]``

        Note:
            Performance On i7-6700HQ (M: Million):

            90ms for building a tree with 10M elements.

            230ms for looking up 10M elements in a tree with 10M elements.

            20ms for 1M element batched update in a tree with 10M elements.

            240ms for 1M element single update in a tree with 10M elements.

        Args:
            size: Number of weight tree leaves.
        """
        self.size = size
        # depth: level of nodes
        self.max_leaf = -np.inf
        self.depth = int(np.ceil(np.log2(self.size))) + 1
        level_sizes_log = np.arange(self.depth - 1, -1, -1)
        self.sizes = np.power(2, level_sizes_log)
        self.offsets = np.concatenate(([0], np.cumsum(self.sizes)))
        self.weights = np.zeros([self.offsets[-1]], dtype=np.float)

    def build(self):
        """
        Build weight tree from leaves
        """
        self.max_leaf = np.max(self.get_leaf_all_weights())
        for i in range(self.depth - 1):
            offset = self.offsets[i]
            level_size = self.sizes[i]
            # data are interleaved, therefore must be -1,2 and not 2,-1
            weight_sum = self.weights[offset: offset + level_size] \
                .reshape(-1, 2).sum(axis=1)

            offset += level_size
            next_level_size = self.sizes[i + 1]
            self.weights[offset: offset + next_level_size] = weight_sum

    def get_weight_sum(self) -> float:
        """
        Returns:
            Total weight sum.
        """
        return self.weights[-1]

    def get_leaf_max(self) -> float:
        """
        Returns:
            Current maximum leaf weight.
        """
        return self.max_leaf

    def get_leaf_all_weights(self) -> np.ndarray:
        """
        Returns:
            Current weights of all leaves, ``np.ndarray`` of shape ``(size)``.
        """
        return self.weights[range(self.sizes[0])]

    def get_leaf_weight(self, index: Union[int, List[int], np.ndarray]) -> Any:
        """
        Get weights of selected leaves.

        Args:
            index: Leaf indexes used to query weights.

        Returns:
            Current weight(s) of selected leaves. If index is scalar, returns
            ``float``, if not, returns ``np.ndarray``.
        """
        if not isinstance(index, np.ndarray):
            index = np.array(index).reshape(-1)
        if np.any(np.array(index) >= self.size):
            raise RuntimeError("Index has elements above buffer size "
                               "boundary!")
        return self.weights[index]

    def find_leaf_index(self, weight: Union[float, List[float], np.ndarray]):
        """
        Find leaf indexes given weight. Weight must be in range
        :math:`[0, weight\\_sum]`

        Args:
            weight: Weight(s) used to query leaf index(es).

        Returns:
            Leaf index(es), if weight is scalar, returns ``int``, if not,
            returns ``np.ndarray``.
        """
        if not isinstance(weight, np.ndarray):
            weight = np.array(weight).reshape(-1)
        index = np.zeros([weight.shape[0]], dtype=np.long)

        # starting from the first child level of root
        for i in range(self.depth - 2, -1, -1):
            offset = self.offsets[i]
            left_wt = self.weights[offset + index * 2]
            # 0 for left and 1 for right
            select = weight > left_wt
            index = index * 2 + select
            weight = weight - left_wt * select
        index = np.clip(index, 0, self.size - 1)

        if index.shape[0] == 1:
            return int(index)
        else:
            return index

    def update_leaf(self, weight: float, index: int):
        """
        Update a single weight tree leaf.

        Args:
            weight: New weight of the leaf.
            index: Leaf index to update, must be in range ``[0, size]``.
        """
        self.max_leaf = max(weight, self.max_leaf)
        self.weights[index] = weight
        value = weight
        comp_value = self.weights[index ^ 0b1]

        for i in range(1, self.depth - 1):
            offset = self.offsets[i]
            index = index // 2
            global_index = index + offset
            comp_global_index = index ^ 0b1 + offset
            value = self.weights[global_index] = value + comp_value
            comp_value = self.weights[comp_global_index]

        self.weights[-1] = value + comp_value

    def update_leaf_batch(self,
                          weights: Union[List[float], np.ndarray],
                          indexes: Union[List[int], np.ndarray]):
        """
        Update weight tree leaves in batch.

        Args:
            weights: New weights of leaves.
            indexes: Leaf indexes to update, must be in range ``[0, size]``.
        """
        if len(weights) != len(indexes):
            raise RuntimeError(
                "Dimension of weights and indexes doesn't match!")
        if np.any(np.array(indexes) >= self.size):
            raise RuntimeError("Index has elements above buffer size boundary!")
        if len(weights) == 0:
            return

        weights = np.array(weights)
        indexes = np.array(indexes)

        self.max_leaf = max(np.max(weights), self.max_leaf)

        needs_update = indexes
        self.weights[indexes] = weights
        for i in range(1, self.depth):
            offset, prev_offset = self.offsets[i], self.offsets[i - 1]
            # O(n) = nlg(n)
            needs_update = np.unique(needs_update // 2)
            tmp = needs_update * 2
            self.weights[offset + needs_update] = \
                self.weights[prev_offset + tmp] + \
                self.weights[prev_offset + tmp + 1]

    def update_all_leaves(self, weights: Union[List[float], np.ndarray]):
        """
        Reset all leaf weights, rebuild weight tree from ground up.

        Args:
            weights: All leaf weights. List or array length should be in range
                ``[0, size]``.
        """
        if len(weights) != self.size:
            raise RuntimeError("Weights size must match buffer size!")
        self.weights[0: len(weights)] = np.array(weights)
        self.build()

    def print_weights(self, precision=2):
        """
        Pretty print the tree, for debug purpose.

        Args:
            precision: Number of digits of weights to print.
        """
        fmt = "{{:.{}f}}".format(precision)
        for i in range(self.depth):
            offset, size = self.offsets[i], self.sizes[i]
            weights = [fmt.format(self.weights[j]) for j in
                       range(offset, offset + size)]
            print(weights)


class PrioritizedBuffer(Buffer):
    def __init__(self, buffer_size, buffer_device="cpu",
                 epsilon=1e-2, alpha=0.6, beta=0.4,
                 beta_increment_per_sampling=0.001, *_, **__):
        """
        Args:
            buffer_size: Maximum buffer size.
            buffer_device: Device where buffer is stored.
            epsilon: A small positive constant used to prevent edge-case
                zero weight transitions from never being visited.
            alpha: Prioritization weight. Used during transition sampling:
                :math:`j \\sim P(j)=p_{j}^{\\alpha} / \
                        \\sum_i p_{i}^{\\alpha}`.
                When ``alpha = 0``, all samples have the same probability
                to be sampled.
                When ``alpha = 1``, all samples are drawn uniformly according
                to their weight.
            beta: Bias correcting weight. When ``beta = 1``, bias introduced
                by prioritized replay will be corrected. Used during
                importance weight calculation:
                :math:`w_j=(N \\cdot P(j))^{-\\beta}/max_i w_i`
            beta_increment_per_sampling:
                Beta increase step size, will gradually increase ``beta`` to 1.
        """
        super(PrioritizedBuffer, self).__init__(buffer_size, buffer_device)
        self.epsilon = epsilon
        self.alpha = alpha
        self.beta = beta
        self.beta_increment_per_sampling = beta_increment_per_sampling
        self.curr_beta = beta
        self.wt_tree = WeightTree(buffer_size)

    def _normalize_priority(self, priority):
        """
        Normalize priority and calculate :math:`p_{j}^{\alpha}`
        """
        return (np.abs(priority) + self.epsilon) ** self.alpha

    def append(self,
               transition: Union[Transition, Dict],
               priority: Union[float, None] = None,
               required_attrs=("state", "action", "next_state",
                               "reward", "terminal")):
        """
        Store a transition object to buffer.

        Args:
            transition: A transition object.
            priority: Priority of transition.
            required_attrs: Required attributes.
        """
        position = super(PrioritizedBuffer, self).append(transition,
                                                         required_attrs)
        if priority is None:
            # the initialization method used in the original essay
            priority = self.wt_tree.get_leaf_max()
        self.wt_tree.update_leaf(self._normalize_priority(priority), position)

    def size(self):
        """
        Returns:
            Length of current buffer.
        """
        return len(self.buffer)

    def clear(self):
        """
        Clear and resets the buffer to its initial state.
        """
        self.buffer.clear()
        self.wt_tree = WeightTree(self.buffer_size)
        self.curr_beta = self.beta

    def update_priority(self, priorities: np.ndarray, indexes: np.ndarray):
        """
        Update priorities of samples.

        Args:
            priorities: New priorities.
            indexes: Indexes of samples, returned by :meth:`sample_batch`
        """
        priorities = self._normalize_priority(priorities)
        self.wt_tree.update_leaf_batch(priorities, indexes)

    def sample_batch(self,
                     batch_size: int,
                     concatenate: bool = True,
                     device: Union[str, t.device] = None,
                     sample_attrs: List[str] = None,
                     additional_concat_attrs: List[str] = None,
                     *_, **__) -> Tuple[Any, Any, np.ndarray, np.ndarray]:
        """
        Sample the most important batch from the prioritized buffer.

        See Also:
             :meth:`.Buffer.sample_batch`

        Args:
            batch_size: A hint size of the result sample. actual sample size
                        depends on your sample method.
            concatenate: Whether concatenate state, action and next_state
                         in dimension 0.
                         If ``True``, for each value in dictionaries of major
                         attributes. and each value of sub attributes, returns
                         a concatenated tensor. Custom Attributes specified in
                         ``additional_concat_attrs`` will also be concatenated.
                         If ``False``, return a list of tensors.
            device:      Device to copy to.
            sample_attrs: If sample_keys is specified, then only specified keys
                         of the transition object will be sampled. You may use
                         ``"*"`` as a wildcard to collect remaining keys.
            additional_concat_attrs: additional custom keys needed to be
                         concatenated,

        Returns:
            Batch size,
            Sampled attribute values in the same order as ``sample_keys``,
            Indexes of samples in the weight tree,
            Importance sampling weight of samples.
        """
        segment_length = self.wt_tree.get_weight_sum() / batch_size
        self.curr_beta = np.min(
            [1., self.curr_beta + self.beta_increment_per_sampling])

        rand_priority = np.random.uniform()
        rand_priority += np.arange(batch_size, dtype=np.float) * segment_length
        rand_priority = np.clip(rand_priority, 0,
                                self.wt_tree.get_weight_sum() - 1e-6)
        index = self.wt_tree.find_leaf_index(rand_priority)

        batch = [self.buffer[idx] for idx in index]
        priority = self.wt_tree.get_leaf_weight(index)

        # calculate importance sampling weight
        sample_probability = priority / self.wt_tree.get_weight_sum()
        is_weight = np.power(len(self.buffer) * sample_probability, -self.beta)
        is_weight /= is_weight.max()

        # post processing
        if device is None:
            device = self.buffer_device
        if sample_attrs is None:
            sample_attrs = batch[0].keys()
        if additional_concat_attrs is None:
            additional_concat_attrs = []

        result = self.post_process_batch(batch, device, concatenate,
                                         sample_attrs, additional_concat_attrs)
        return batch_size, result, index, is_weight



import time


class Counter:
    def __init__(self, start=0, step=1):
        self._count = start
        self._start = start
        self._step = step

    def count(self):
        """
        Move counter forward by ``step``
        """
        self._count += self._step

    def get(self):
        """
        Get the internal number of counter.
        """
        return self._count

    def reset(self):
        """
        Reset the counter.
        """
        self._count = self._start

    def __lt__(self, other):
        return self._count < other

    def __gt__(self, other):
        return self._count > other

    def __le__(self, other):
        return self._count <= other

    def __ge__(self, other):
        return self._count >= other

    def __eq__(self, other):
        return self._count == other

    def __repr__(self):
        return "%d" % self._count


class Switch:
    def __init__(self, state: bool = False):
        """
        Args:
            state: Internal state, ``True`` for on, ``False`` for off.
        """
        self._on = state

    def flip(self):
        """
        Inverse the internal state.
        """
        self._on = not self._on

    def get(self) -> bool:
        """
        Returns:
            state of switch.
        """
        return self._on

    def on(self):
        """
        Set to on.
        """
        self._on = True

    def off(self):
        """
        Set to off.
        """
        self._on = False


class Trigger(Switch):
    def get(self):
        """
        Get the state of trigger, will also set trigger to off.

        Returns:
            state of trigger.
        """
        on = self._on
        if self._on:
            self._on = False
        return on


class Timer:
    def __init__(self):
        self._last = time.time()

    def begin(self):
        """
        Begin timing.
        """
        self._last = time.time()

    def end(self):
        """
        Returns:
            Curent time difference since last ``begin()``
        """
        return time.time() - self._last


class Object:
    """
    An generic object class, which stores a dictionary internally, and you
    can access and set its keys by accessing and seting attributes of the
    object.

    Attributes:
        data: Internal dictionary.
    """
    def __init__(self, data=None):
        if data is None:
            data = {}
        super(Object, self).__setattr__("data", data)

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)

    def call(self, *args, **kwargs):
        """
        the implementation of ``Object.__call__``, override it to
        customize call behavior.
        """
        pass

    def __getattr__(self, item):
        # This function will be called if python cannot find an attribute
        # Note: in order to make Object pickable, we must raise AttributeError
        # when looking for special methods, such that when pickler is looking
        # up a non-existing __getstate__ function etc, this class will
        # not return a None value because self.attr(item) will return None.
        if isinstance(item, str) and item[:2] == item[-2:] == '__':
            # skip non-existing special method lookups
            raise AttributeError(item)
        return self.attr(item)

    def __getitem__(self, item):
        return self.attr(item)

    def __setattr__(self, key, value):
        if key != "data" and key != "attr" and key != "call":
            self.attr(key, value)
        elif key == "call":
            super(Object, self).__setattr__(key, value)
        elif key == "data":
            if isinstance(value, dict):
                super(Object, self).__setattr__(key, value)
            else:
                raise RuntimeError("The data attribute must be a dictionary.")
        else:
            raise RuntimeError("You should not set the attr property of an "
                               "Object. Please Override it in a sub class.")

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def attr(self, item, change=None):
        if change is not None:
            self.data[item] = change
        return self.data.get(item, None)



from time import sleep
from datetime import timedelta
from threading import Thread, local
from typing import Union, List, Tuple, Dict, Any, Callable
from torch.distributed import rpc

import torch.distributed as dist
import torch.distributed.distributed_c10d as dist_c10d

from .election import ElectionGroupStableRpc
from .role_dispatcher import RoleDispatcherElection

WORLD = None  # type: Union[None, World]


def _copy_doc(from_func):
    """
    Used by collective group to copy documents from torch.
    """
    import io
    import sys

    def _decorator(func):
        if 'sphinx' in sys.modules:
            src_doc = from_func.__doc__
            lines = io.StringIO(src_doc)
            # remove the group line
            src_doc = "".join([line for line in lines if "group" not in line])
            func.__doc__ = src_doc
        return func
    return _decorator


def _exec_role(role):
    # role thread executor.
    local.role = role
    role.on_init()
    role.main()
    role.on_stop()


def _rpc_call_method(method, inst, *args, **kwargs):
    # Call a method function of a class, ``inst`` is a class instance
    return method(inst, *args, **kwargs)


def _rpc_call_remote_method(method, inst_rref, *args, **kwargs):
    # Call a method function of a class,
    # ``inst_rref`` is a class instance wrapped by ``RRef``
    return method(inst_rref.local_value(), *args, **kwargs)


def _rpc_call_nn_module(nn_module, *args, **kwargs):
    return nn_module(*args, **kwargs)


def _rpc_get_remote_paired_value(group_name, key):
    # TODO: dump other paired maps of the same group
    global WORLD
    paired_map = WORLD.groups[group_name].group_paired_map
    if key in paired_map:
        return paired_map[key]
    else:
        raise RpcException("""
            Failed to find key {} in the paired value map of:
            Process(rank={}, name={}) from Group {}
        """.format(key, WORLD.current_rank, WORLD.current_name, group_name))


def _world_singleton(cls, *args, **kwargs):
    # Decorator used to makesure that there is only one world instance.
    def _world_singleton_wrapper():
        global WORLD
        if WORLD is None:
            WORLD = cls(*args, **kwargs)
        else:
            raise RuntimeError("World could only be initialized once!")

    return _world_singleton_wrapper


def get_cur_rank():
    """
    Returns:
        Current real process rank.
    """
    if WORLD is None:
        raise RuntimeError("Distributed environment not initialized!")
    return WORLD.current_rank


def get_cur_role():
    """
    Returns:
        Current thread role (thread is a thread of current real process).
    """
    return local.role


class RpcException(Exception):
    """
    Rpc exception class.
    """
    pass


@_world_singleton
class World:
    """
    The distributed world.
    """
    def __init__(self,
                 world_size: int,
                 current_rank: int,
                 roles: Dict[str, Tuple[type, int]],
                 init_method: str = "tcp://localhost:9100",
                 rpc_timeout: int = 60,
                 rpc_threads: int = 4,
                 rpc_role_dispatcher: Any = None
                 ):
        """
        Args:
            world_size:   Size of distributed world.
            current_rank: A unique rank of current process.
            roles: A list of roles executed by all processes.
            init_method:  Backend initialization method.
            rpc_timeout:  Global rpc call timeout in seconds.
            rpc_threads:  Rpc recv/send thread num.
            rpc_role_dispatcher: Rpc role dispatch, by default it is
                :class:`~machin.parallel.distributed.\
RoleDispatcherElection` and uses :class:`machin.parallel.\
distributed.ElectionGroupStableRpc` as its internal election implementation.
        """
        self.world_size = world_size
        self.role_dict = roles
        # Maps role Tuple[str, int] to threads
        self.role_threads = {}

        self.current_rank = current_rank
        self.ranks = [i for i in range(world_size)]
        self.real_names = ["{}".format(i) for i in range(world_size)]
        self.groups = {}
        if rpc_role_dispatcher is not None:
            self.rpc_role_dispatcher = rpc_role_dispatcher
        else:
            role_names = list(roles.keys())
            role_nums = [val[1] for val in roles.values()]
            self.rpc_role_dispatcher = RoleDispatcherElection(
                current_rank, world_size,
                role_names, role_nums,
                ElectionGroupStableRpc(
                    name="global",
                    member_ranks=self.ranks,
                    rank=current_rank,
                    timeout=rpc_timeout
                )
            )

        # "<rank-number>" is used as the unique name.
        rpc.init_rpc("{}".format(self.current_rank),
                     rank=current_rank,
                     world_size=world_size,
                     rpc_backend_options=rpc.ProcessGroupRpcBackendOptions(
                         init_method=init_method,
                         num_send_recv_threads=rpc_threads,
                         rpc_timeout=timedelta(seconds=rpc_timeout)
                     ))

        # Start role dispatching.
        self.rpc_role_dispatcher.start()
        while True:
            self.rpc_role_dispatcher.get_role_update_cond().wait()
            for role in self.rpc_role_dispatcher.get_roles():
                if role not in self.role_threads:
                    role_class = self.role_dict[role[0]][0]
                    role_thread = Thread(target=_exec_role,
                                         args=(role_class(role[1]),))
                    role_thread.start()
                    self.role_threads[role] = role_thread

    def create_collective_group(self,
                                group_name: str,
                                ranks: List[int],
                                timeout: Any = dist.default_pg_timeout,
                                backend: Any = None):
        """
        Create a sub process group for collective communications.

        Args:
            group_name: A unique group name.
            ranks: Ranks of involved processes.
            timeout: Timeout of operations in the new group.
            backend: New group backend.
        Returns:
            A ``Group`` with type ``Group.COLLECTIVE``
        """
        if group_name in self.groups:
            raise RuntimeError("Group {} already existed!".format(group_name))
        ranks = sorted(ranks)
        group = CollectiveGroup(dist.new_group(ranks, timeout, backend),
                                ranks.index(self.current_rank))
        return group

    def create_rpc_group(self, group_name: str, roles: List[Any]):
        """
        Create a sub process group for rpc calls.

        Args:
            group_name: A unique group name.
            roles: Roles of involved processes.

        Returns:
            A ``Group`` with type ``Group.RPC``
        """
        if group_name in self.groups:
            raise RuntimeError("Group {} already existed!".format(group_name))
        group = RpcGroup(group_name, roles)
        self.groups[group_name] = group
        return group

    def get_group(self, group_name: str):
        """
        Get group with name ``group_name``

        Args:
            group_name: Group name

        Returns:
            Target group
        """
        return self.groups.get(group_name, None)

    def __reduce__(self):
        raise RuntimeError("World is not pickable, create it per process!")


class CollectiveGroup:
    """
    A thin wrapper of collective communication primitives of
    ``torch.distributed``, the only difference is that ``irecv``
    now supports to recv from any src
    """
    def __init__(self, group, current_relative_rank):
        """
        Do not do it your self, use :meth:`~machin.parallel\
.distributed.World.create_collective_group` instead.
        """
        self.group = group
        self.current_rank = current_relative_rank
        self.destroyed = False

    @_copy_doc(dist.send)
    def send(self, tensor, dst, tag=0):
        return dist.send(tensor, dst, self.group, tag)

    @_copy_doc(dist.recv)
    def recv(self, tensor, src=None, tag=0):
        return dist.recv(tensor, src, self.group, tag)

    @_copy_doc(dist.isend)
    def isend(self, tensor, dst, tag=0):
        return dist.isend(tensor, dst, self.group, tag)

    @_copy_doc(dist.irecv)
    def irecv(self, tensor, src=None, tag=0):
        # pylint: disable=protected-access

        # Original irecv doesn't support recv from any
        # but original recv does. They are essentially
        # the same except recv have a wait() call
        dist_c10d._check_single_tensor(tensor, "tensor")
        if dist_c10d._rank_not_in_group(self.group):
            return -1

        if self.group == dist_c10d.GroupMember.WORLD:
            dist_c10d._check_default_pg()
            pg = dist_c10d._default_pg
        else:
            pg = self.group

        if src is None:
            work = pg.recv_anysource([tensor], tag)
            src_rank = work.source_rank()
            if self.group == dist_c10d.GroupMember.WORLD:
                return src_rank
            else:
                return dist_c10d._get_global_rank(pg, src_rank)
        else:
            if self.group == dist_c10d.GroupMember.WORLD:
                pg.recv([tensor], src, tag).wait()
            else:
                group_src_rank = dist_c10d._get_group_rank(pg, src)
                pg.recv([tensor], group_src_rank, tag).wait()
            return src

    @_copy_doc(dist.broadcast)
    def broadcast(self, tensor, async_op=False):
        return dist.broadcast(tensor, self.current_rank, self.group, async_op)

    @_copy_doc(dist.all_reduce)
    def all_reduce(self, tensor, op=dist.ReduceOp.SUM, async_op=False):
        return dist.all_reduce(tensor, op, self.group, async_op)

    @_copy_doc(dist.reduce)
    def reduce(self, tensor, dst, op=dist.ReduceOp.SUM, async_op=False):
        return dist.reduce(tensor, dst, op, self.group, async_op)

    @_copy_doc(dist.all_gather)
    def all_gather(self, tensor_list, tensor, async_op=False):
        return dist.all_gather(tensor_list, tensor, self.group, async_op)

    @_copy_doc(dist.gather)
    def gather(self, tensor, gather_list, dst=0, async_op=False):
        return dist.gather(tensor, gather_list, dst, self.group, async_op)

    @_copy_doc(dist.scatter)
    def scatter(self, tensor, scatter_list=None, src=0, async_op=False):
        return dist.scatter(tensor, scatter_list, src, self.group, async_op)

    @_copy_doc(dist.barrier)
    def barrier(self, async_op=False):
        return dist.barrier(self.group, async_op)

    def destroy(self):
        """
        Destroy this collective communication group.
        """
        if not self.destroyed:
            dist.destroy_process_group(self.group)
            self.destroyed = True

    def size(self):
        """
        Returns:
            collective group size.
        """
        return dist.get_world_size(self.group)

    def __reduce__(self):
        raise RuntimeError("Group is not pickable, create it per process!")

    def __del__(self):
        self.destroy()


RoleHandle = Union[str, Tuple[str, int]]


class RpcGroup:
    def __init__(self, group_name, group_roles):
        self.group_name = group_name
        self.group_roles = group_roles
        self.group_paired_map = {}
        self.destroyed = False

    def rpc_sync(self,
                 to: RoleHandle, func: Callable,
                 timeout=-1, retry=True, require_in_group=True,
                 args=(), kwargs=None):
        """
        Synchronous rpc call.

        Args:
            to: Role name, e.g.: "some_role:10".
            func: Some function.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        Returns:
            Function results.
        """
        return self._rpc_normal_call(rpc.rpc_sync, to, func,
                                     timeout, retry, require_in_group,
                                     args, kwargs)

    def rpc_async(self,
                  to: RoleHandle, func: Callable,
                  timeout=-1, retry=True, require_in_group=True,
                  args=(), kwargs=None):
        """
        Asynchronous rpc call.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            func: Some function.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        Returns:
            A rpc future object you can call ``.wait()`` on.
        """
        return self._rpc_normal_call(rpc.rpc_async, to, func,
                                     timeout, retry, require_in_group,
                                     args, kwargs)

    def rpc_remote(self,
                   to: RoleHandle, func: Callable,
                   timeout=-1, retry=True, require_in_group=True,
                   args=(), kwargs=None):
        """
        Remote rpc call.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            func: Some function.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        Returns:
            A ``RRef`` object.
        """
        return self._rpc_normal_call(rpc.remote, to, func,
                                     timeout, retry, require_in_group,
                                     args, kwargs)

    def rpc_register_paired(self, name: Any, value: Any):
        """
        Register a paired value to current process group.

        Args:
            name: A key which uniquely identifies this value in this group.
                 The name only needs to be unique for this value in this
                 group.

            value: Value to be registered.
        """
        self.group_paired_map[name] = value

    def rpc_get_paired(self, target: RoleHandle, name: Any,
                       timeout=-1, retry=True):
        """
        Args:
            target: Role name "some_role:10" or role tuple ("some_role", 10).
            name: Name of the paired value to get.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.

        Returns:
            An RRef to the paired value.

        Raises:
            :class:`.RpcException` if not found.
        """
        # TODO: add timeout
        del timeout

        while True:
            if not self.is_member(target):
                raise RuntimeError("Target is not a member of the group.")
            rpc_target = self._get_real_name(target)
            try:
                return rpc.remote(rpc_target,
                                  _rpc_get_remote_paired_value,
                                  args=(self.group_name, name))
            except RuntimeError:
                WORLD.rpc_role_dispatcher.notify_failure(
                    self._parse_role(target)
                )
                if not retry:
                    break
                sleep(0.1)

    def rpc_paired_class_sync(self,
                              to: RoleHandle, cls_method: Callable, name: Any,
                              timeout=-1, retry=True, require_in_group=True,
                              args=(), kwargs=None):
        """
        Call the specified ``cls_method`` on ``to`` using ``name`` to find
        the class instance.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            cls_method: Class method, e.g.:``some_class.some_method``
            name: Class instance name.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        See Also:
            :meth:`.RpcGroup.rpc_sync`
        """
        return self._rpc_paired_class_call(rpc.rpc_sync, to, cls_method, name,
                                           timeout, retry, require_in_group,
                                           args, kwargs)

    def rpc_paired_class_async(self,
                               to: RoleHandle, cls_method: Callable, name: Any,
                               timeout=-1, retry=True, require_in_group=True,
                               args=(), kwargs=None):
        """
        Call the specified ``cls_method`` on ``to`` using ``name`` to find
        the class instance.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            cls_method: Class method, e.g.:``some_class.some_method``
            name: Class instance name.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        See Also:
            :meth:`.RpcGroup.rpc_async`
        """
        return self._rpc_paired_class_call(rpc.rpc_async, to, cls_method, name,
                                           timeout, retry, require_in_group,
                                           args, kwargs)

    def rpc_paired_class_remote(self,
                                to: RoleHandle, cls_method: Callable, name: Any,
                                timeout=-1, retry=True, require_in_group=True,
                                args=(), kwargs=None):
        """
        Call the specified ``cls_method`` on ``to`` using ``name`` to find
        the class instance.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            cls_method: Class method, e.g.:``some_class.some_method``
            name: Class instance name.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        See Also:
            :meth:`.RpcGroup.rpc_remote`
        """
        return self._rpc_paired_class_call(rpc.remote, to, cls_method, name,
                                           timeout, retry, require_in_group,
                                           args, kwargs)

    def rpc_paired_nn_module_sync(self,
                                  to: RoleHandle, name: Any,
                                  timeout=-1, retry=True, require_in_group=True,
                                  args=(), kwargs=None):
        """
        Run the forward pass on ``to`` using ``name`` to find
        the model instance.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            name: Model instance name.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        See Also:
            :meth:`.RpcGroup.rpc_sync`
        """
        return self._rpc_paired_nn_module_call(rpc.rpc_sync, to, name,
                                               timeout, retry, require_in_group,
                                               args, kwargs)

    def rpc_paired_nn_module_async(self,
                                   to: RoleHandle, name: Any,
                                   timeout=-1, retry=True,
                                   require_in_group=True,
                                   args=(), kwargs=None):
        """
        Run the forward pass on ``to`` using ``name`` to find
        the model instance.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            name: Model instance name.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        See Also:
            :meth:`.RpcGroup.rpc_async`
        """
        return self._rpc_paired_nn_module_call(rpc.rpc_async, to, name,
                                               timeout, retry, require_in_group,
                                               args, kwargs)

    def rpc_paired_nn_module_remote(self,
                                    to: RoleHandle, name: Any,
                                    timeout=-1, retry=True,
                                    require_in_group=True,
                                    args=(), kwargs=None):
        """
        Run the forward pass on ``to`` using ``name`` to find
        the model instance.

        Args:
            to: Role name "some_role:10" or role tuple ("some_role", 10).
            name: Model instance name.
            timeout: Call timeout.
            retry: Whether to retry until success after timeout.
            require_in_group: Whether require the call target to be in group.
            args: Arguments.
            kwargs: Key arguments.

        See Also:
            :meth:`.RpcGroup.rpc_remote`
        """
        return self._rpc_paired_nn_module_call(rpc.remote, to, name,
                                               timeout, retry, require_in_group,
                                               args, kwargs)

    def _rpc_normal_call(self, rpc_method, to, func,
                         timeout, retry, require_in_group,
                         args, kwargs):
        """
        TODO: add timeout.
        """
        del timeout

        while True:
            if require_in_group and not self.is_member(to):
                raise RuntimeError("RPC target is not a member of group.")
            rpc_to = self._get_real_name(to)
            args = list(args)
            try:
                if hasattr(func, "__self__"):
                    # instance rref, instance might be builtin or user-defined
                    args = [func, func.__self__] + args
                    return rpc_method(rpc_to, _rpc_call_method, args=args,
                                      kwargs=kwargs)
                else:
                    return rpc_method(rpc_to, func, args=args, kwargs=kwargs)
            except RuntimeError:
                WORLD.rpc_role_dispatcher.notify_failure(
                    self._parse_role(to)
                )
                if not retry:
                    break
                sleep(0.1)

    def _rpc_paired_class_call(self, rpc_method, to, cls_method, name,
                               timeout, retry, require_in_group,
                               args, kwargs):
        """
        TODO: add timeout.
        """
        del timeout

        while True:
            if require_in_group and not self.is_member(to):
                raise RuntimeError("RPC target is not a member of group.")
            rpc_to = self._get_real_name(to)
            if hasattr(cls_method, "__self__"):
                inst_rref = self.rpc_get_paired(to, name)
                args = tuple([cls_method, inst_rref] + list(args))
                try:
                    return rpc_method(rpc_to, _rpc_call_remote_method,
                                      args=args, kwargs=kwargs)
                except RuntimeError:
                    WORLD.rpc_role_dispatcher.notify_failure(
                        self._parse_role(to)
                    )
                    if not retry:
                        break
                    sleep(0.1)
            else:
                raise RuntimeError("Method does not belong to "
                                   "any registered paired class")

    def _rpc_paired_nn_module_call(self, rpc_method, to, name,
                                   timeout, retry, require_in_group,
                                   args, kwargs):
        """
        TODO: add timeout.
        """
        del timeout

        while True:
            if require_in_group and not self.is_member(to):
                raise RuntimeError("RPC target is not a member of group.")
            rpc_to = self._get_real_name(to)
            nnm_rref = self.rpc_get_paired(to, name)
            args = tuple([nnm_rref] + list(args))
            try:
                return rpc_method(rpc_to, _rpc_call_nn_module,
                                  args=args, kwargs=kwargs)
            except RuntimeError:
                WORLD.rpc_role_dispatcher.notify_failure(
                    self._parse_role(to)
                )
                if not retry:
                    break
                sleep(0.1)

    def destroy(self):
        """
        Destroy the rpc group.
        """
        if not self.destroyed:
            if len(WORLD.groups) == 1:
                rpc.shutdown(True)
                WORLD.groups.clear()
            else:
                WORLD.groups.pop(self.group_name)
            self.destroyed = True

    def size(self):
        """
        Get the number of roles in group.
        """
        return len(self.group_roles)

    def is_member(self, role: str) -> bool:
        """
        Check whether ``role`` is a group member.

        Args:
            role: A string like "some_role:10"
        """
        return self._parse_role(role) in self.group_roles

    def get_group_members(self) -> List[Tuple[str, int]]:
        """
        Returns:
            A list of group members (roles).
        """
        return self.group_roles

    @staticmethod
    def get_cur_role():
        """
        Returns:
            Current role of the executor thread.
        """
        return get_cur_role()

    @classmethod
    def _get_real_name(cls, role: str) -> str:
        # get the real rpc process name used in rpc api call
        role = cls._parse_role(role)
        return str(WORLD.rpc_role_dispatcher.get_rank(role))

    @staticmethod
    def _parse_role(role: RoleHandle) -> Tuple:
        if (isinstance(role, tuple) and len(role) == 2 and
                isinstance(role[0], str) and isinstance(role[1], int)):
            return role
        # parse a role string "some_role:10" to tuple ("some_role", 10)
        role = list(role.split(':'))
        role[1] = int(role[1])
        role = tuple(role)
        return role

    def __reduce__(self):
        raise RuntimeError("Group is not pickable, create it per process!")

    def __del__(self):
        self.destroy()



from . import \
    checker, conf, helper_classes, learning_rate, \
    logging, media, prepare, save_env, tensor_board, \
    visualize

__all__ = ["checker", "conf", "helper_classes", "learning_rate",
           "logging", "media", "prepare", "save_env", "tensor_board",
           "visualize"]



"""
This module is the place for all learning rate functions, currently, only
manual learning rate changing according to global steps is implemented,.
"""
from typing import List, Tuple
from logging import Logger


def gen_learning_rate_func(lr_map: List[Tuple[int, int]],
                           logger: Logger = None):
    """
    Example::

        from torch.optim.lr_scheduler import LambdaLR

        # 0 <= step < 200000, lr=1e-3, 200000 <= step, lr=3e-4
        lr_func = gen_learning_rate_func([(0, 1e-3), (200000, 3e-4)],)
        lr_sch = LambdaLR(optimizer, lr_func)

    Args:
        lr_map: A 2d learning rate map, the first element of each row is step.
            the second is learning rate.
        logger: A logger to log current learning rate

    Returns:
        A learning rate generation function with signature `lr_gen(step)->lr`,
        accepts int and returns float. use it in your pytorch lr scheduler.
    """
    def learning_rate_func(step):
        for i in range(len(lr_map) - 1):
            if lr_map[i][0] <= step < lr_map[i + 1][0]:
                if logger is not None:
                    logger.info("Current learning rate:{}".format(lr_map[i][1]))
                return lr_map[i][1]
        if logger is not None:
            logger.info("Current learning rate:{}".format(lr_map[-1][1]))
        return lr_map[-1][1]

    return learning_rate_func



class RoleBase(object):
    """
    Base of all roles, derive from this class to implement your own roles.

    :ivar NAME: A unique name of this role.
    """
    NAME = ""

    def __init__(self, index):
        """
        Args:
            index: worker index of this role, e.g.: ``"some_role:10"`` should
                have ``10`` as index.
        """
        self.role_index = index

    def on_init(self):
        """
        :meth:`on_init` is executed before the process executes your event loop.
        """
        pass

    def on_stop(self):
        """
        :meth:`on_init` is executed after the process stops your event loop and
        exit
        """
        pass

    def main(self):
        """
        :meth:`main` Is the main event loop of the role
        """
        pass

    def __str__(self):
        return self.NAME + ":" + str(self.role_index)
